/* crc.c - crc-32 report, w/directory recursion, by date

    crc makes a pretty good substitute for ux' ls and find, or dos' dir/s;
    if you shut off the crc calculation with -c, it'll run down a tree
    very quickly to glob a filename; you can ask for date and size as well
    as either Zip or cksum crc values; from VMS it borrows timestamp
    filtering, either /before or /since a given time (not both);
    crc does not show or select on ownership or mode;

20070830.35.11:06:34 - timestamps
20080820_34_16:26:46 - filename globbing, improved recursion, zip crc
20081214_51_21:28:30 - -k (kat) -x (exec) commands, on each filename
                    nb:  -k reports each file normally before the kat,
                        -xkat does a naked kat of each file; your choice
*/

#include <dirent.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>

#define TRUE    1
#define FALSE   0

    char pthseg[4] = "/";

    int     asc = 0;
    int     szz = 0;
    int     nll = 0;
    int     bnr = 1;
    int     crc = 1;
    int     cmp = 0;
    int     csm = 1;
    int     dte = 0;
    int     kat = 0;
    int     siz = 0;
    int     tbl = 0;
    int     rcr = 1;
    int     rtm = 0;
    int     flx = 0;
    int     dos = 0;
    int     vrb = 0;
    int     cnt = 0;
    int     drs = 0;
    int     grn = 0;
    double  grs = 0.;

    char    **flt = NULL;
    char    **blk = NULL;

    char    cmd[1024] = "";

    unsigned long crctab[256];

int main(int argc, char *argv[])
{
    int     i = 0;
    int     j = 0;
    int     s = 0;
    int     o = 0;
    int     c = 0;
    int     x = 0;
    unsigned long x1 = 0;
    unsigned long x2 = 0;
    time_t  t = 0;
    struct  tm ttim;
    char    nm[256] = ".";
    char    dr[256] = "";
    char    fn[256] = "";
    char    rg[256] = "";
    char    nw[40] = "";
    char    st[40] = "";
    char    nd[40] = "";
    struct  stat fs;
    FILE   *fp = NULL;

    unsigned long recur(char *nm, int o, time_t t);
    char *now(char *n);
    void fltr(char *fn);
    void crc_init();
    void fltadd(char *b);
    void blkadd(char *b);
    void fnm_split (char *fs, char *dr, char *nm);

    if (argc < 2) {
        printf(" crc - recursive directory listing, with crc-32 report; usage:\n\n");
        printf(" crc <-sw> <dir/file*.\?\?\?>\n\n");
        printf("    optional toggle switches (1=on by default):\n");
        printf("        -= = compare two files, report mis-match (%d)\n",cmp);
        printf("        -b = binary-read (%d)\n",bnr);
        printf("        -c = report crc-32 (%d)\n",crc);
        printf("        -d = report modification date (%d)\n",dte);
        printf("        -f<fn> = file of filename selectors\n");
        printf("        -g = grand total only:  runtime, count, size\n");
        printf("        -k = report & kat each filename found (%d)\n",kat);
        printf("        -n = report nulls & junk (%d)\n",nll);
        printf("        -p = set path separator = \\ (for dos)\n");
        printf("        -r = directory recursion (%d)\n",rcr);
        printf("        -s = report file size (%d)\n",siz);
        printf("        -t = dump crc-32 table (%d)\n",tbl);
        printf("        -v = verbose (%d)\n",vrb);
        printf("        -x<cmd> = execute <cmd> with each filename as parameter\n");
        printf("        -z = use Zip algorithm instead of cksum (%d)\n",csm);
        printf("    optional filters, using wildcard globbing:\n");
        printf("        -f<name> = file of extension filters\n");
        printf("        -*<glob> = inclusion filter (selector)\n");
        printf("        -#<glob> = exclusion filter\n");
        printf("    optional timestamp, optional sign:\n");
        printf("        -@<ts> or '@' or\n");
        printf("        @<ts> = report files before or after <ts>;\n");
        printf("          if <ts> is not supplied, the default is 'today';\n");
        printf("          ts format: [+|-]yyyymmdd[.ww.hh.mm.ss]\n\n");
        printf("        by default (+), files in the directory tree\n");
        printf("        which are newer than the timestamp are reported;\n");
        printf("        if the sign is negative, older files are reported.\n");
        printf("    notes:\n");
        printf("      <dir/file> is the only parameter, everything else is a\n");
        printf("        switch; depending on your shell's behavior, wildcards\n");
        printf("        (or any filespec) might be expanded in unexpected ways;\n");
        printf("        but you can prevent this by using quotes, and crc will still\n");
        printf("        handle wildcards, e.g.:  \"./*.c\";\n");
        printf("      <dir/> will be traversed recursively, unless you suppress it\n");
        printf("        with '-r';\n");
        printf("      if you use -*<glob> or -#<glob> filters, specify only a\n");
        printf("        directory name as the parameter to scan, not dir/file,\n");
        printf("        otherwise there will be some funny conflicts;\n");
        printf("      the switch characters '*' and '#' are not included\n");
        printf("        in the filters - if you want a leading '*', add another;\n");
        printf("      you may specify as many -* or -# filters as you need;\n");
        printf("      cross-platform comparison of text files will probably\n");
        printf("        require '-b' to turn off binary reading of cr/lf.\n");
        printf("\n    Wed Dec 24 14:26:31 2008 359 - e.roosa\n");
        exit(0);
    }

    now(st);

    for (i = 1; i < argc; i++) {
        switch (argv[i][0]) {
            case '-':
                for (j = 1; j < strlen(argv[i]); j++) {
                    switch (c = tolower(argv[i][j])) {
                        case '=':   cmp = !cmp; break;
                        case 'b':   bnr = !bnr; break;
                        case 'c':   crc = !crc; break;
                        case 'd':   dte = !dte; break;
                        case 'f':   strcpy(nm,&argv[i][j+1]);
                                    j = strlen(argv[i]);
                                    fltr(nm);
                                    break;
                        case 'g':   rtm = grn = 1; break;
                        case 'n':   nll = !nll; break;
                        case 'p':   pthseg[0] = '\\'; break;
                        case 'r':   rcr = !rcr; break;
                        case 's':   siz = !siz; break;
                        case 't':   tbl = !tbl;
                                    if (tbl && !crctab[1]) crc_init();
                                    break;
                        case 'v':   vrb = dte = rtm = siz = 1; break;
                        case 'k':   strcpy(cmd,"kat");
                                    kat = 1;
                                    break;
                        case 'x':   strcpy(cmd,&argv[i][j+1]);
                                    j = strlen(argv[i]);
                                    break;
                        case 'z':   csm = !csm; break;
                        case '#':   blkadd(&argv[i][j+1]);
                                    j = strlen(argv[i]);
                                    break;
                        case '*':   flx = 1;
                                    fltadd(&argv[i][j+1]);
                                    j = strlen(argv[i]);
                                    break;
                        case '@':   strcpy(rg,&argv[i][j]);
                                    j = strlen(argv[i]);
                                    break;
                        default:    fprintf(stderr,"!! %c:  unrecognized switch\n",c);
                                    break;
                    }
                }
                break;
            case '@':
                strcpy(rg,argv[i]);
                break;
            default: {
                if (!t) {
                    time(&t);
                    ttim = *localtime(&t);
                    if (!nw[0]) strcpy(nw,"19700101.00.00:00:01");
                    ttim.tm_sec  = atoi(&nw[18]);     nw[17] = '\0';
                    ttim.tm_min  = atoi(&nw[15]);     nw[14] = '\0';
                    ttim.tm_hour = atoi(&nw[12]);     nw[8]  = '\0';
                    ttim.tm_mday = atoi(&nw[6]);      nw[6]  = '\0';
                    ttim.tm_mon  = atoi(&nw[4])  - 1; nw[4]  = '\0';
                    ttim.tm_year = atoi(nw) - 1900;
                    t = mktime(&ttim);
                }
                strcpy(fn,argv[i]);
                //printf("fn %s %s\n",fn,nw);
                if (s = stat(fn,&fs)) {
                    fnm_split(fn,dr,nm);
                    fltadd(nm);
                    recur(dr,o,t);
                } else {
                    if ((cmp) && (x1)) {
                        x2 = recur(fn,o,t);
                        if (x1 != x2) {
                            fnm_split(fn,dr,nm);
                            printf("  !!\n  != %s\n  !!\n",nm);
                            x = 3;
                        }
                    } else {
                        //fnm_split(fn,dr,nm);
                        x1 = recur(fn,o,t);
                    }
                }
                break;
            }
        }
        if (rg[0]) {
            if (rg[1]) {
                if (stat(&rg[1],&fs)) {
                    strcpy(nw,&rg[1]);
                } else {
                    fp = fopen(&rg[1],"r");
                    fgets(nw,sizeof(nw),fp);
                    fclose(fp);
                }
            } else {
                strncpy(nw,st,8);
            }
            if (nw[0] == '-') {
                o = 1;
                strcpy(rg,nw);
                strcpy(nw,&rg[1]);
            }
            if (nw[0] == '+') {
                o = 0;
                strcpy(rg,nw);
                strcpy(nw,&rg[1]);
            }
            rg[0] = '\0';
        }
    }
//    if (rtm) fprintf(stderr,"\n start:  %s\n   end:  %s\n count:  %d\n  size:  %.0f\n",st,now(nd),cnt,grs);
    if (rtm) {
        fprintf(stderr,"-------------------- ----------\n");
        fprintf(stderr,"%s:%10d files, %d director%s\n%s %10.0f bytes.\n",st,cnt,drs,((drs==1)?"y":"ies"),now(nd),grs);
    }
    exit(x);
}

// use the glob lists to choose or discard each filename passed:
int fltx(char *fn) {
    int i = 0;
    int s = 0;
    int fnm_match(char *pattern, char *fs, int flags);
    // fnm_match:  glob filespec; 1=match, 0=no match

    if (blk) for (i = 0; blk[i]; i++) {
        if (fnm_match(blk[i],fn,0)) return(0);
        //if (strstr(fn,blk[i])) return(0);
    }
    if (flt) { for (i = 0; flt[i]; i++) {
        if (fnm_match(flt[i],fn,0)) return(1);
    }} else return(1);
    return(0);
}

// add a file of filters, 1 per line:
void fltr(char *fn) {
    char    b[1024] = "*";
    FILE   *fp = NULL;
    void fltadd(char *b);

    if (fp = fopen(fn,"r")) {
        while (fgets(&b[1],1023,fp)) {
            b[strlen(b)-1] = '\0';
            fltadd(b);
        }
        fclose(fp);
    } else {
        fprintf(stderr,"!! can't open extension filter %s\n",fn);
    }
}

// add a filter (glob pattern) to the list;
//  if we match any one of these, we like the filename;
void fltadd(char *b) {
    static  int n = 0;
    int     s = 0;
    char   *c = NULL;

    if (!(n%64)) flt = realloc(flt,(n+65)*sizeof(char *));
    s = strlen(b);
    //fprintf(stderr," %s %d %d\n",b,n,s);
    if (s > 0) {
        c = calloc(1,s+2);
        strcpy(c,b);
        flt[n++] = c;
        flt[n] = NULL;
        //fprintf(stderr," %s %d %p\n",c,n,flt[n-1]);
    }
    //fprintf(stderr,"%s %d\n",flt[n-1],n);
}

// add a blocking pattern to the list;
//  if we match any one of these, toss the filename;
void blkadd(char *b) {
    static  int n = 0;
    int     s = 0;
    char   *c = NULL;

    if (!(n%64)) blk = realloc(blk,(n+65)*sizeof(char *));
    s = strlen(b);
    //fprintf(stderr," %s %d %d\n",b,n,s);
    if (s > 0) {
        c = calloc(1,s+2);
        strcpy(c,b);
        blk[n++] = c;
        blk[n] = NULL;
        //fprintf(stderr," %s %d %p\n",c,n,blk[n-1]);
    }
    //fprintf(stderr,"%s %d\n",blk[n-1],n);
}

int fnm_match (char *pattern, char *fs, int flags) {
    static char dr[256] = "";
    static char nm[256] = "";
    void fnm_split (char *fs, char *dr, char *nm);
    int wc_match(char *pTameText, char *pWildText, int bCaseSensitive, char cAltTerminator);

    if (strlen(fs) > 255) return(0);
    fnm_split(fs, dr, nm);
    return(wc_match(nm,pattern,0,'\0'));
}

void fnm_split (char *fs, char *dr, char *nm) {
    int k = 0;

    if (fs != dr) strcpy(dr,fs);
    for (k = strlen(dr); k > 0; k--) {
        if ((dr[k] == '\\')
        ||  (dr[k] == '/')) {
            dr[k++] = '\0';
            strcpy(nm,&dr[k]);
            break;
        } else {
        if ((dr[k] == ':')
        ||  (dr[k] == ']')) {
            k++;
            strcpy(nm,&dr[k]);
            dr[k] = '\0';
            break;
        }}
    }
    if (k==0) {
        strcpy(nm,dr);
        strcpy(dr,".");
    }
    if (nm[0]=='\0') {
        strcpy(nm,"*");
    }
}

// Matching Wildcards: An Algorithm
// by Kirk J. Krauss
//      http://www.ddj.com
// This function compares text strings, one of which can have wildcards ('*').
//
int wc_match(char *pTameText, char *pWildText, int bCaseSensitive, char cAltTerminator) {
                       // A string without wildcards
                                        // A (potentially) corresponding string with wildcards
                                                         // By default, match on 'X' vs 'x'
                                                                             // For function names, for example, 
                                                                             // you can stop at the first '('

        int  bMatch = TRUE;
        char *pAfterLastWild = NULL;  // The location after the last '*', if we've encountered one
        char *pAfterLastTame = NULL;  // The location in the tame string, from which we started after last wildcard
        char t, w;

        // Walk the text strings one character at a time.
        while (1)
        {
               t = *pTameText;
               w = *pWildText;

               // How do you match a unique text string?
               if (!t || t == cAltTerminator)
               {
                       // Easy: unique up on it!
                       if (!w || w == cAltTerminator)
                       {
                              break;                                 // "x" matches "x"
                       }
                       else if (w == '*')
                       {
                              pWildText++;
                              continue;                          // "x*" matches "x" or "xy"
                       }
                       else if (pAfterLastTame)
                       {
                              if (!(*pAfterLastTame) || *pAfterLastTame == cAltTerminator)
                              {
                                      bMatch = FALSE;
                                      break;
                              }
                              pTameText = pAfterLastTame++;
                              pWildText = pAfterLastWild;
                              continue;
                       }

                       bMatch = FALSE;
                       break;                                         // "x" doesn't match "xy"
               }
               else
               {
                       if (!bCaseSensitive)
                       {
                              // Lowercase the characters to be compared.
                              if (t >= 'A' && t <= 'Z')
                              {
                                      t += ('a' - 'A');
                              }

                              if (w >= 'A' && w <= 'Z')
                              {
                                      w += ('a' - 'A');
                              }
                       }

                       // How do you match a tame text string?
                       if (t != w)
                       {
                              // The tame way: unique up on it!
                              if (w == '*')
                              {
                                      pAfterLastWild = ++pWildText;
                                      pAfterLastTame = pTameText;
                                      continue;                          // "*y" matches "xy"
                              }
                              else if (pAfterLastWild)
                              {
                                      pWildText = pAfterLastWild;
                                      w = *pWildText;

                                      if (!w || w == cAltTerminator)
                                      {
                                             break;                          // "*" matches "x"
                                      }
                                      else
                                      {
                                             if (!bCaseSensitive && w >= 'A' && w <= 'Z')
                                             {
                                                     w += ('a' - 'A');
                                             }
                                             if (t == w)
                                             {
                                                     pWildText++;
                                             }
                                      }
                                      pTameText++;
                                      continue;                          // "*sip*" matches "mississippi"
                              }
                              else
                              {
                                      bMatch = FALSE;
                                      break;                                 // "x" doesn't match "y"
                              }
                       }
               }

               pTameText++;
               pWildText++;
        }

        return bMatch;
}

char *now(char *n) {
    time_t          psec;
    struct tm      *ptim;

    time(&psec);
    ptim = localtime(&psec);
    sprintf(n,"%04d%02d%02d.%02d.%02d:%02d:%02d"
        ,ptim->tm_year+1900
        ,ptim->tm_mon+1
        ,ptim->tm_mday
        ,(ptim->tm_yday+7-ptim->tm_wday)/7+1
        ,ptim->tm_hour
        ,ptim->tm_min
        ,ptim->tm_sec);
    n[20] = '\0';
    return(n);
}

unsigned long recur(char *nm, int old, time_t nsec) {
    DIR *dr;
    struct  dirent  *dn;
    struct  stat s;
    int r;
    int newer = 0;
    char fn[512] = "";
    char ds[512] = "";
    char now[40] = "";
    unsigned long rpt(char *fn, int old, time_t f, time_t c, int s);

    strcpy(ds,nm);
    r = stat(ds,&s);
    if (!S_ISDIR(s.st_mode)) {
        if (!r) return(rpt(ds,old,s.st_mtime,nsec,s.st_size));
        return(0);
    }
    if (dr = opendir(ds)) {
        strcat(ds,pthseg);
        while ((dn = readdir(dr)) != NULL) {
            if (strcmp(dn->d_name,"..") && strcmp(dn->d_name,".")) {
                strcpy(fn,ds);
                strcat(fn,dn->d_name);
                r = stat(fn,&s);
                if (S_ISDIR(s.st_mode)) {
                    if (rcr) recur(fn,old,nsec);
                } else {
                    rpt(fn,old,s.st_mtime,nsec,s.st_size);
                }
            }
        }
        closedir(dr);
        drs++;
    } else {
        fprintf(stderr,"!! directory open failed:  %s (%d)\n",ds,errno);
    }
    return(0);
}

unsigned long rpt(char *fn, int old, time_t ftim, time_t ctim, int s) {
    struct tm *tim;
    int newer = (difftime(ftim,ctim) > 0);
    unsigned long x = 0;
    unsigned long scan(char *fn,int s);
    int fltx(char *fn);
    char cm[1024] = "";

    if (fltx(fn)) {
    if (old) newer = !newer;
    if (newer) {
        if ((s) && (cmd[0]) && !(kat)) {
            //printf("\n\n######### -->%s<--\n\n\n",fn);
            strcpy(cm,cmd);
            strcat(cm," ");
            strcat(cm,fn);
            system(cm);
        } else {
            if (asc) {
                fprintf(stderr,"!! non-ascii in %s\n",fn);
                asc = 0;
            }
            if (szz) {
                fprintf(stderr,"!! file size error in %s\n",fn);
                szz = 0;
            }
            if (rtm) {
                grs += s;
                cnt++;
            }
            if (!grn) {
            if (dte) {
                tim = localtime(&ftim);
                printf("%4d%02d%02d.%02d.%02d:%02d:%02d ",
                    (tim->tm_year + 1900),
                    (tim->tm_mon+1),
                    tim->tm_mday,
                    (tim->tm_yday+7-tim->tm_wday)/7+1,
                    tim->tm_hour,
                    tim->tm_min,
                    tim->tm_sec
                    );
            }
            if (siz) {
                printf("%10d ",s);
            }
            if (crc) {
                x = scan(fn,s);
                printf("%08x ",x);
                if (vrb) printf("%010u ",x);
            }
            printf("%s\n",fn);
            }
            if (kat) {
                strcpy(cm,cmd);
                strcat(cm," ");
                strcat(cm,fn);
                system(cm);
            }
        }
    }
    }
    return(x);
}

unsigned long scan(char *fn, int s) {
    int     i = 0;
    int     c = 0;              /* char read */
    unsigned char *b = NULL;    /* buffer for summing */
    char    bf[8092] = "";      /* text file line buffer */
    char    fm[4] = "r";        /* file-open mode ('rb'?) */
    FILE   *fp = NULL;          /* read file ptr */
    unsigned long x = 0;
    unsigned long z = 0;        /* char count */
    unsigned long l = 0;        /* char count */
    unsigned long crc32(unsigned char* p, unsigned long len);
    unsigned long cksm(unsigned char* p, unsigned long len);

    if (bnr) fm[1] = 'b';
    if ((fp = fopen(fn, fm)) == NULL) return(0);
    b = realloc(b,s+4);
    if (bnr) {
        while (!feof(fp) && ((c = fgetc(fp)) != EOF)) {
            b[z++] = c;
        }
    } else if (nll) {
        while (!feof(fp) && ((c = fgetc(fp)) != EOF)) {
            if ((c < 9) || (c > 126) || ((c > 13) && (c < 32))) asc = 1;
            if (c != 13) b[z++] = c;
            l++;
        }
        if (s != l) szz = 1;
    } else {
        while (!feof(fp) && ((c = fgetc(fp)) != EOF)) {
            if (c != 13) b[z++] = c;
        }
        /*
        while (fgets(bf,8091,fp)) {
            for (i = 0; (bf[i]); i++) {
                c = bf[i];
                b[z++] = c;
            }
        }*/
    }
    fclose(fp);
    if (b) {
        b[z] = '\0'; b[z+1] = '\0'; b[z+2] = '\0'; b[z+3] = '\0';
        if (csm) {
            x = crc32(b,z);
        } else {
            x = cksm(b,z);
        }
    }
    return(x);
}

/*
    there are 2 algorithms here, one that matches cksum, one that matches Zip;
    by default, we use the Zip calculation on the assumption that we might be
    looking at stuff stored in .zip or .jar files for comparison;
    both of these appear in cygwin, gnu, et.al.  nothin' new here. . .
*/

const int order = 32;
const unsigned long polynom = 0x4c11db7;
const int direct = 1;
const unsigned long crcinit = 0xffffffff;
const unsigned long crcxor = 0xffffffff;
const int refin = 1;
const int refout = 1;

unsigned long crcmask;
unsigned long crchighbit;
unsigned long crcinit_direct;
unsigned long crcinit_nondirect;

unsigned long crctab[256];

unsigned long reflect (unsigned long crc, int bitnum)
{
    // reflects the lower 'bitnum' bits of 'crc'

    unsigned long i, j=1, crcout=0;

    for (i=(unsigned long)1<<(bitnum-1); i; i>>=1) {
        if (crc & i) crcout|=j;
        j <<= 1;
    }
    return (crcout);
}

void generate_crc_table()
{
    // make CRC lookup table used by table algorithms

    int i, j;
    unsigned long bit, crc;

    for (i=0; i<256; i++) {

        crc=(unsigned long)i;
        if (refin) crc = reflect(crc, 8);
        crc <<= order-8;

        for (j=0; j<8; j++) {
            bit = crc & crchighbit;
            crc <<= 1;
            if (bit) crc ^= polynom;
        }                       

        if (refin) crc = reflect(crc, order);
        crc &= crcmask;
        crctab[i]= crc;
    }
}

void crc_init()
{
    int i, j;
    unsigned long bit, crc;

    // at first, compute constant bit masks for whole CRC and CRC high bit

    crcmask = ((((unsigned long)1<<(order-1))-1)<<1)|1;
    crchighbit = (unsigned long)1<<(order-1);

    // compute missing initial CRC value

    if (!direct) {
        crcinit_nondirect = crcinit;
        crc = crcinit;
        for (i=0; i<order; i++) {
            bit = crc & crchighbit;
            crc <<= 1;
            if (bit) crc ^= polynom;
        }
        crc &= crcmask;
        crcinit_direct = crc;
    } else {
        crcinit_direct = crcinit;
        crc = crcinit;
        for (i=0; i<order; i++) {
            bit = crc & 1;
            if (bit) crc ^= polynom;
            crc >>= 1;
            if (bit) crc |= crchighbit;
        }
        crcinit_nondirect = crc;
    }

    generate_crc_table();

    if (tbl) {
        for (i = 0; i < 32; i++) {
            for (j = 0; j < 8; j++) {
                fprintf(stderr," 0x%08x,",crctab[i*8+j]);
            }
            fprintf(stderr,"\n");
        }
        fprintf(stderr,"\n");
    }

/*
    printf("Parameters:\n");
    printf("\n");
    printf(" polynom             :  0x%x\n", polynom);
    printf(" order               :  %d\n", order);
    printf(" crcinit             :  0x%x direct, 0x%x nondirect\n", crcinit_direct, crcinit_nondirect);
    printf(" crcxor              :  0x%x\n", crcxor);
    printf(" refin               :  %d\n", refin);
    printf(" refout              :  %d\n", refout);
    printf("\n");
*/
}

unsigned long crcbitbybit(unsigned char* p, unsigned long len)
{
    // bit by bit algorithm with augmented zero bytes.
    // does not use lookup table, suited for polynom orders between 1...32.

    unsigned long i, j, c, bit;
    unsigned long crc = crcinit_nondirect;

    for (i=0; i<len; i++) {
        c = (unsigned long)*p++;
        if (refin) c = reflect(c, 8);

        for (j=0x80; j; j>>=1) {
            bit = crc & crchighbit;
            crc<<= 1;
            if (c & j) crc|= 1;
            if (bit) crc^= polynom;
        }
    }

    for (i=0; i<order; i++) {
        bit = crc & crchighbit;
        crc<<= 1;
       if (bit) crc^= polynom;
    }

    if (refout) crc=reflect(crc, order);
    crc^= crcxor;
    crc&= crcmask;

    return(crc);
}

unsigned long crc32(unsigned char* p, unsigned long len)
{
    unsigned long crc = 0;

    // fast lookup table algorithm without augmented zero bytes, e.g. used in pkzip.
    // only usable with polynom orders of 8, 16, 24 or 32.

    if (!crctab[1]) crc_init();

    //return(crcbitbybit(p,len));

    crc = crcinit_direct;

    if (refin) crc = reflect(crc, order);

    if (!refin)
        while (len--) crc = (crc << 8) ^ crctab[ ((crc >> (order-8)) & 0xff) ^ *p++];
    else
        while (len--) crc = (crc >> 8) ^ crctab[ (crc & 0xff) ^ *p++];

    if (refout^refin) crc = reflect(crc, order);

    crc^= crcxor;
    crc&= crcmask;

    return(crc);
}

#define QUOTIENT    0x04C11DB7

unsigned long cksmtab[256];

void cksmtab_init()
{
    int i,j;
    unsigned int crc;

    for (i = 0; i < 256; i++) {
        crc = i << 24;
        for (j = 0; j < 8; j++) {
            if (crc & 0x80000000)
                crc = (crc << 1) ^ QUOTIENT;
            else
                crc = crc << 1;
        }
        cksmtab[i] = crc;
    }

    if (tbl) {
        for (i = 0; i < 32; i++) {
            for (j = 0; j < 8; j++) {
                fprintf(stderr," 0x%08x,",cksmtab[i*8+j]);
            }
            fprintf(stderr,"\n");
        }
        fprintf(stderr,"\n");
    }
}

unsigned long cksm(unsigned char *b, unsigned long z)
{
    int i;
    int c;
    unsigned long x1 = 0;       /* checksum */
    unsigned long x2 = 0;       /* checksum */
    unsigned long  x = 0;       /* checksum */
             long  y = 0;       /*  -decimal */

    if (!cksmtab[1]) cksmtab_init();

    for (i = 0; i < z; i++) {
        c = b[i] & 0xff;
        x1 = (x >> 24) ^ c;
        x2 = x << 8;
        x = x2 ^ cksmtab[x1 & 0xFF];
    }
    y = z;
    while (y > 0) {
        x1 = (x >> 24) ^ y;
        x2 = x << 8;
        x = x2 ^ cksmtab[x1 & 0xFF];
        y >>= 8;
    }
    x = ~x & 0xFFFFFFFF;
    return(x);
}

