/* kat.c - kat, w/prefix

    concatenate all the files named on the command-line to stdout; also:

    if argv == "stdin"  copy from stdin (i.e., pipe);
    if argv == "#n"     pad each line to n blanks; if n < 0, trim whitespace;
    if argv == "<text"  write 'text' as the next line (use quotes, for the '<');
    if argv == ">fn"    redirect output to file 'fn' (only needed by vms);
    if argv == "@n"     chop left n columns;
    if argv == "="      be like uniq;
    if argv == "~"      unlike uniq, print only dupes;
    if argv == ".c"     subst '\n' for 'c', eg:  '.;' for dos path pretty-print;

    The usual output of a VMS text editor is a 'sequential/variable/carriage-control'
    text file.  But a lot of programs, especially u**x or c-based, write
    something VMS calls 'stream-lf'.  They look the same, and print the same,
    but as input to many programs they do not behave the same.  Conversely,
    many programs, especially mainframe-based, expect fixed-length file input,
    and for text on VMS, this is a very unusual file format.

    So kat does seq-var output when asked to, and it can pad text records to a
    fixed length.  It acts as much like a u**x pipe as I needed it to, which is
    to say, all input filenames are prefixed with '<' and the output filename
    is prefixed with '>'.

    What it does on VMS, kat does on u**x as well, although it is not as essential
    there.  Kat's not trying to be a drop-in 'cat' replacement, but the padding
    and the text-inclusion might sometimes be useful.

    Mon Aug 27 15:10:35 2007 239, v2:

    a - fiddle chars, prefix '.' followed by a list of chars, each of which will
        become '\n' in the piped output, e.g., "path |kat stdin .;"

    b - @n

    c - uniq, parameter '='; input must be sorted to work right;

    d - un-uniq:  parameter '~' signals print only dupes; sort first;

    e - dimensions:  param '$' - show lines/cols;

    Fri Dec 12 12:35:11 2008 347:

        wide chars ignored, not supported (but at least they print);

*/

    const char *BUILD = "Fri Dec 12 12:34:52 2008 347 - earlroosa@yahoo.com";

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

    char **lb = NULL;   // tail lines (data)
    int  el = 0;        // tail lines (req)
    int  le = 0;        // tail lines (cur)
    int  wc = 0;        // wide(unicode) chars

int main(int argc, char *argv[])
{
    int i = 0;  // iterator
    int j = 0;  // iterator
    int a = 0;  // print only from this col
    int z = 32767;  // truncate at this col
    int c = 0;  // input char
    int d = 0;  // dimension report (lin/wid)
    int e = 0;  // token end
    FILE *f = NULL; // binary file, for nl control
    int h = 0;  // head line count
    int l = 0;  // line count
    int n = 0;  // input newline?
    int o = 0;  // offset, i + (mod[tabs])
    int p = 0;  // pad count
    int r = 0;  // arg counter
    int s = 0;  // token start
    int t = 0;  // token, for uniq
    int u = 0;  // uniq?
    int w = 0;  // width
    int x = 0;  // binary output to control newline
    char lt[100] = "";  // artificial line terminators (PATH?)
    char fn[256] = "";  // binary output filename
    char rd[256] = "";  // input filename
    char b[32002] = ""; // output buffer
    char q[32002] = ""; // uniq buffer
    char m[32002] = ""; // uniq match-token buffer

    void kaput(char *t,int i,int x);
    void kaputs();

    if (argc < 2) {
        strcpy(b,"\n kat [-sw] [\"stdin\"] [\"<text\"] inFile [\"<text2\"] [inFile2 . . .] [-nOutFile]"); puts(b);
        strcpy(b,"\n    "); strcat(b,BUILD); puts(b);
        exit(0);
    }
    for (r = 1; r < argc; r++) {
      if (argv[r][0] == '-') {
        switch (argv[r][1]) {
            case '?':
            case 'H':
                strcpy(b," kat [options] [\"stdin\"] [\"<text\"] infile [\"<text2\"] [infile2 . . .]\n"); puts(b);
                strcpy(b,"    read a pipe or a file, and write to stdout (pipe or file);"); puts(b);
                strcpy(b,"    filenames (w/no prefixes) will be opened and copied in order;"); puts(b);
                strcpy(b,"    a bare string, or a string enclosed in quotes, with a prefix of '<'"); puts(b);
                strcpy(b,"    is assumed to be a one-line inclusion, prepended to whatever follows;"); puts(b);
                strcpy(b,"    options:"); puts(b);
                strcpy(b,"      '-#n':  pad the output to a fixed length n using spaces;"); puts(b);
                strcpy(b,"      '-#-1': trim the output of whitespace;"); puts(b);
                strcpy(b,"      '-=':   do 'uniq' on the input;"); puts(b);
                strcpy(b,"      '-~':   do un-uniq, print only dupes;"); puts(b);
                strcpy(b,"      '-$':   report lines/maxcols;"); puts(b);
                strcpy(b,"      '-.;':  turn ';' into '\\n' (e.g.);"); puts(b);
                strcpy(b,"      '-[n':  chop left n chars;"); puts(b);
                strcpy(b,"      '-]n':  truncate at n chars;"); puts(b);
                strcpy(b,"      '-hn':  head, n lines;"); puts(b);
                strcpy(b,"      '-tn':  tail, n lines;"); puts(b);
                strcpy(b,"      '-n<file>':  newline control:  add \\r in ux, drop it in windoze;\n"); puts(b);
                strcpy(b,"    "); strcat(b,BUILD); puts(b);
                exit(0);
                break;
            case '#':
                p = atoi(&argv[r][2]);
                break;
            case '[':
                a = atoi(&argv[r][2]);
                break;
            case ']':
                z = atoi(&argv[r][2]);
                break;
            case 'h':
                h = atoi(&argv[r][2]);
                break;
            case 't':
                el = atoi(&argv[r][2]);
                lb = calloc(el+1,sizeof(char *));
                //for (i = 0; i < el; i++) printf("%p ",lb[i]);
                //printf("\n");
                break;
            case '=':
                u = 1;
                t = atoi(&argv[r][2]);
                break;
            case '~':
                u = -1;
                t = atoi(&argv[r][2]);
                break;
            case '.':
                strcpy(lt,&argv[r][2]);
                break;
            case '$':
                d = 1;
                break;
            case 'n':
                x = 1;
                if ((f = freopen(&argv[r][2],"wb",stdout)) == NULL) {
                    strcpy(b,"!! can't reopen ");
                    strcat(b,&argv[r][2]);
                    exit(2);
                }
                break;
            default: break;
        }
      }
    }
    r = 1;
    while (argc > r) {
        if (argv[r][0] == '<') {
            strcpy(b,&argv[r][1]);
            i = strlen(b);
            while (i < p) b[i++] = ' ';
            kaput(b,i,x);
        } else if (argv[r][0] != '-') {
            if (strcmp("stdin",argv[r])) {
                strcpy(rd,argv[r]);
                if ((f = freopen(argv[r],"r",stdin)) == NULL) {
                    strcpy(rd,argv[r]);
                    strcat(rd,".lis");
                if ((f = freopen(rd,"r",stdin)) == NULL) {
                    strcpy(rd,argv[r]);
                    strcat(rd,".txt");
                if ((f = freopen(rd,"r",stdin)) == NULL) {
                    break;
                }}}
            }
            wc = w = l = n = 0;
            for (i = 0;;) {
                c = fgetc(stdin);
                if ((c > 253) && (i == 0)) { wc = 1; continue; }
                if (c == '\0') { continue; }
                if (c == '\n') {
                    n = fgetc(stdin);
                    if (n != '\r') ungetc(n,stdin);
                }
                for (j = 0; lt[j]; j++) if (c==lt[j]) c = '\n';
                if ((i >= 32000) || (c == '\n') || ((i) && (c < 0))) {
                    l++;
                    if (d) { if (w < i) w = i; }
                    if (p<0) { i--; while (i>0) { if (b[i] <= ' ') { b[i] = '\0'; i--; } else { break; } } i++; }
                    while (o < p) { b[i++] = ' '; o++; }
                    b[i] = '\0';
                    if (i > a) i = a;
                    if (u) {
                        if (t) {
                            s = 0;
                            for (j = 1; j < t; j++) {
                                while (b[s] >= ' ') s++;
                                if (b[s]) s++;
                            }
                            strcpy(m,&b[s]);
                            e = 0;
                            while (m[e] >= ' ') e++;
                            m[e] = '\0';
                        } else {
                            strcpy(m,b);
                        }
                        if ((u > 0) && !strcmp(q,m)) i = -1;
                        if ((u < 0) &&  strcmp(q,m)) i = -1;
                        strcpy(q,m);
                    }
                    if (i >= 0) kaput(b,i,x);
                    i = o = 0;
                }
                if (c < 0) break;
                if (n < 0) break;
                if ((h) && (l >= h)) break;
                if (o > z) continue;
                if ((c != '\n') && (c != '\r')) b[i++] = (char)c;
                if (c == '\t') o += 8 - (o%8); else o++;
            }
            if (el) kaputs();
            if (d) fprintf(stderr,"%d lines %d maxcols %s\n",l,w,argv[r]);
        }
        r++;
    }
    exit(0);
}

void kaput(char *t,int i, int x) {
    int o = 0;
    int e = 0;
    char *b = t;

    while (o < i) if (b[i] == '\t') o += 8 - (o%8); else o++;
    b = &t[o];
    e = strlen(b);

#if __WIN32__
    if (!x) b[e++] = '\r';
#else
    if (x)  b[e++] = '\r';
#endif

    b[e++] = '\n';
    b[e++] = '\0';

    if (el == 0) {
        fwrite(b,strlen(b),1,stdout);
        return;
    }
    lb[le%el] = (char *)realloc(lb[le%el],strlen(b)+2);
    strcpy(lb[le%el],b);
    //printf("%p %s\n",lb[le%el],lb[le%el]);
    le++;
}

void kaputs() {
    int i = 0;
    for (i = 0; i < el; i++) {
        fwrite(lb[le%el],strlen(lb[le%el]),1,stdout);
        le++;
    }
}
